import numpy as np
from functools import reduce
import matplotlib.pyplot as plt
from copy import deepcopy
import pandas as pd
import datetime as dt
from scipy.stats import norm

import ray
import psutil

def gen_probs(nstates):
    x = np.random.random(size=nstates)
    x /= np.sum(x)
    return pd.Series({s: x[s] for s in range(nstates)}).sort_index()

def gen_data(n_obs, bpolicy, P_MDP, expected_rewards, noise = 0):
    Data = []
    nstates, nactions = bpolicy.shape
    for _ in range(n_obs):
        state=np.random.choice(nstates, size=1)[0]
        action = np.random.choice(np.arange(nactions), p=bpolicy[state,:])
        state_ = np.random.choice(np.arange(nstates), p=P_MDP.loc[(state,action),:])
        reward = expected_rewards[state] + np.random.normal(scale=noise)
        Data.append((state,action,reward,state_))
        state = state_
    return Data

def online_bootstrap(Data, Phi, gamma, alpha0, eta, policy, bpolicy, B, freq):
    CIs = {'Q': [], 'SE': []}
    p = Phi.shape[1]
    Theta = np.zeros([B+1,2*p])
    Thetabar = np.zeros([B+1,p])

    j=0
    for i in range(len(Data)):
        j+=1
        alpha_t = alpha0*j**(-eta)
        W = np.concatenate([[1], np.random.exponential(size=B)])
        state,action,reward,state_ = Data[i]
        rho_t = policy[state,action] / bpolicy[state,action]
        phi = Phi[state,:]
        phi_ = Phi[state_,:]
        At = rho_t * np.outer(phi, phi - gamma*phi_)
        bt = rho_t * reward * phi
        Mt = np.outer(phi, phi)
        A1t = np.row_stack([np.column_stack([Mt, At]), np.column_stack([-At.T, np.zeros(shape=(p,p))])])
        b1t = np.concatenate([bt, np.zeros(p)])
        Theta += alpha_t*np.diag(W) @ (b1t[np.newaxis,:] - Theta@A1t.T)
        Thetabar = ((j-1)*Thetabar + Theta[:,p:])/j
        if i > 0 and i % freq == 0:
            thetabar = Thetabar[0,:]
            ThetabarW = Thetabar[1:,:]
            values = Phi@Thetabar.T
            value0 = values[:,0]
            valueW = values[:,1:]
            CI_Q = value0[:,np.newaxis] + np.quantile(value0[:,np.newaxis] - valueW, [0.025,0.975], axis=1).T
            SE = np.sqrt(np.diag(Phi @ np.cov(thetabar[np.newaxis,:] - ThetabarW, rowvar=False) @ Phi.T))
            CI_SE = value0[:,np.newaxis] + 1.96*np.outer(SE, [-1,1])
            CI_SE = np.insert(CI_SE,1,value0,axis=1)
            CI_Q = np.insert(CI_Q,1,value0,axis=1)
            CIs['Q'].append(CI_Q)
            CIs['SE'].append(CI_SE)
    return CIs

def offline_bootstrap(Data, Phi, gamma, alpha0, eta, policy, bpolicy, B, freq):
    CIs = {'Q': [], 'SE': []}
    p = Phi.shape[1]
    Theta = np.zeros([B+1,2*p])
    Thetabar = np.zeros([B+1,p])
    n_obs = len(Data)
    samples = [np.arange(n_obs)] + [np.random.choice(n_obs,size=n_obs,replace=True) for _ in range(B)]

    js=np.zeros(B+1).astype(int)
    for i in range(len(Data)):
        for b in range(B):
            js[b]+=1
            alpha_t = alpha0 * js[b]**(-eta)
            state,action,reward,state_ = Data[samples[b][i]]
            rho_t = policy[state,action] / bpolicy[state,action]
            phi = Phi[state,:]
            phi_ = Phi[state_,:]
            At = rho_t * np.outer(phi, phi - gamma*phi_)
            bt = rho_t * reward * phi
            Mt = np.outer(phi, phi)
            A1t = np.row_stack([np.column_stack([Mt, At]), np.column_stack([-At.T, np.zeros(shape=(p,p))])])
            b1t = np.concatenate([bt, np.zeros(p)])
            Theta[b,:] += alpha_t*(b1t - A1t@Theta[b,:])
            Thetabar[b,:] = ((js[b]-1)*Thetabar[b,:] + Theta[b,p:])/js[b]
        if i > 0 and i % freq == 0:
            thetabar = Thetabar[0,:]
            ThetabarW = Thetabar[1:,:]
            values = Phi@Thetabar.T
            value0 = values[:,0]
            valueW = values[:,1:]
            CI_Q = value0[:,np.newaxis] + np.quantile(value0[:,np.newaxis] - valueW, [0.025,0.975], axis=1).T
            SE = np.sqrt(np.diag(Phi @ np.cov(thetabar[np.newaxis,:] - ThetabarW, rowvar=False) @ Phi.T))
            CI_SE = value0[:,np.newaxis] + 1.96*np.outer(SE, [-1,1])
            CI_SE = np.insert(CI_SE,1,value0,axis=1)
            CI_Q = np.insert(CI_Q,1,value0,axis=1)
            CIs['Q'].append(CI_Q)
            CIs['SE'].append(CI_SE)
    return CIs

def flatten(list):
    return [item for sublist in list for item in sublist]

def lstd_save(Data, Phi, gamma, policy_target):
    num_states, num_actions = policy_target.shape
    Data_arr = np.array(Data)
    n = Data_arr.shape[0]
    states = Data_arr[:,0].astype(int)
    actions = Data_arr[:,1].astype(int)
    rewards = Data_arr[:,2]
    states_ = Data_arr[:,3].astype(int)

    Xi = np.matmul(np.eye(num_actions)[actions,:,None], Phi[states,None,:]).reshape(n,-1)
    U = np.matmul(policy_target[states_,:,None], Phi[states_,None,:]).reshape(n,-1)
    y = rewards
    Sigma = 1/n * Xi.T@(Xi - gamma*U)
    Sigma_inv = np.linalg.pinv(Sigma)
    beta = 1/n*np.linalg.multi_dot([Sigma_inv, Xi.T, y])
    betamat = beta.reshape(num_actions,-1)
    Qarr = Phi@betamat.T
    values = (Qarr * policy_target).sum(axis=1)

    tmp = np.square(y + gamma*(Phi[states_]*(policy_target[states_,:] @ betamat)).sum(axis=1) - (Phi[states,:]*betamat[actions,:]).sum(axis=1))
    Omega = 1/n*sum([np.outer(Xi[i,:],Xi[i,:])*tmp[i] for i in range(n) ])
    U0 = np.matmul(policy_target[:,:,None], Phi[:,None,:]).reshape(num_states,-1)
    sigma2 = np.array([np.linalg.multi_dot([U0[i,:], Sigma_inv, Omega, Sigma_inv.T, U0[i,:].T]) for i in range(num_states)])
    sigmahat = np.sqrt(1/n * sigma2)
    return values[:,np.newaxis] + np.outer(sigmahat, np.array([-1,0,1])*norm.ppf(0.975))

@ray.remote
def run_bootstrap(seed, n_obs, P_MDP, expected_rewards, policy, bpolicy, Phi, gamma, alpha0, eta, B, freq):
    np.random.seed(seed)
    if seed % 50 == 0:
        now = dt.datetime.now()
        print (seed, now.strftime("%Y-%m-%d %H:%M:%S"))
    Data = gen_data(n_obs, bpolicy, P_MDP, expected_rewards, noise = 0)
    CIs_online = online_bootstrap(Data, Phi, gamma, alpha0, eta, policy, bpolicy, B, freq)
    CIs_offline = offline_bootstrap(Data, Phi, gamma, alpha0, eta, policy, bpolicy, B, freq)
    CIs_save = [lstd_save(Data[:i], Phi, gamma, policy) for i in np.arange(freq,n_obs+freq,freq)]
    return {'online': CIs_online, 'offline': CIs_offline, 'save': CIs_save}

if __name__ == "__main__":
    np.random.seed(2022)
    nstates = 50
    nactions = 5
    p = 10
    b0 = np.random.choice(10, size=p)
    Phi = np.random.normal(size=(nstates, p))
    values0 = Phi@b0

    P_MDP = {s: pd.DataFrame.from_dict({a:gen_probs(nstates) for a in range(nactions)}, orient='index') for s in range(nstates)}
    P_MDP = pd.concat(P_MDP.values(), keys = P_MDP.keys())

    policy = np.row_stack([np.eye(nactions)[np.random.choice(nactions,size=1),:] for _ in range(nstates)])
    random_policy = np.ones([nstates, nactions]) / nactions
    epsilon = 0.2
    bpolicy = (1-epsilon)*policy + epsilon*random_policy

    gamma = 0.99

    expected_rewards = np.zeros(nstates)
    for s in range(nstates):
        expected_rewards[s] = values0[s] - gamma*sum(P_MDP.loc[(s,a),s_]*policy[s,a]*values0[s_] for s_ in range(nstates) for a in range(nactions))

    n_obs = int(1e4)
    alpha0 = 0.1
    eta = 2/3
    B = 200
    freq = 100

    #widths
    Data = gen_data(n_obs, bpolicy, P_MDP, expected_rewards, noise = 0)
    CIs_online = online_bootstrap(Data, Phi, gamma, alpha0, eta, policy, bpolicy, B, freq)
    CIs_offline = offline_bootstrap(Data, Phi, gamma, alpha0, eta, policy, bpolicy, B, freq)
    CIs_save = [lstd_save(Data[:i], Phi, gamma, policy) for i in np.arange(freq,n_obs+freq,freq)]

    dim = 4
    CIs_online_SE = np.array([x[dim,:] for x in CIs_online['SE']])
    CIs_offline_SE = np.array([x[dim,:] for x in CIs_offline['SE']])
    CIs_save_dim = np.array([x[dim,:] for x in CIs_save])

    CIs_online_SE_width = CIs_online_SE[:,2] - CIs_online_SE[:,0]
    CIs_offline_SE_width = CIs_offline_SE[:,2] - CIs_offline_SE[:,0]
    CIs_save_width = CIs_save_dim[:,2] - CIs_save_dim[:,0]

    x = np.arange(len(CIs_online_SE_width))*freq
    plt.rcParams['figure.figsize'] = (16,9)
    plt.rcParams['figure.facecolor'] = 'white'
    plt.plot(x[20:],CIs_online_SE_width[20:], color='blue', label='online bootstrap', linewidth=3)
    plt.plot(x[20:],CIs_offline_SE_width[20:], color='red', label='offline bootstrap', linewidth=3)
    plt.plot(x[20:],CIs_save_width[20:-1], color='green', label='SAVE (LSTD-Q)', linewidth=3)
    plt.xlabel('Number of observations', fontsize='xx-large')
    plt.ylabel('CI Width', fontsize='xx-large')
    plt.xticks(fontsize=20)
    plt.yticks(fontsize=20)
    plt.legend(fontsize='xx-large')
    plt.savefig('offpolicy_compare_widths.png', bbox_inches='tight')
    plt.close()

    #coverage
    niter = 200
    args = {'n_obs': n_obs, 'P_MDP': P_MDP, 'expected_rewards': expected_rewards, 'policy': policy,
            'bpolicy': bpolicy, 'Phi': Phi, 'gamma': gamma, 'alpha0': alpha0, 'eta': eta, 'B': B, 'freq': freq}

    num_cpus = psutil.cpu_count(logical=False)
    ray.init(num_cpus=num_cpus, ignore_reinit_error=True)
    res = ray.get([run_bootstrap.remote(seed, **args) for seed in range(niter)])

    online = [r['online'] for r in res]
    offline = [r['offline'] for r in res]
    save = [r['save'] for r in res]

    dim = 9
    cov_online_SE = np.array([np.array([(values0[dim] > CIs[dim,0]) & (values0[dim] < CIs[dim,2]) for CIs in r['SE']]) for r in online]).mean(axis=0)
    cov_offline_SE = np.array([np.array([(values0[dim] > CIs[dim,0]) & (values0[dim] < CIs[dim,2]) for CIs in r['SE']]) for r in offline]).mean(axis=0)
    cov_save = np.array([np.array([(values0[dim] > CIs[dim,0]) & (values0[dim] < CIs[dim,2]) for CIs in r]) for r in save]).mean(axis=0)
    x = np.arange(len(cov_online_SE))*freq

    plt.rcParams['figure.figsize'] = (16,9)
    plt.rcParams['figure.facecolor'] = 'white'
    plt.plot(x,cov_online_SE,color='blue',label='online bootstrap',linewidth=3)
    plt.plot(x,cov_offline_SE,color='red',label='offline bootstrap',linewidth=3)
    plt.plot(x,cov_save[:-1], color='green',label='SAVE (LSTD-Q)',linewidth=3)
    plt.axhline(0.95, linestyle='--', color='black', linewidth=3)
    plt.legend(fontsize='xx-large')
    plt.xlabel('Number of observations', fontsize='xx-large')
    plt.ylabel('Empirical Coverage Probability', fontsize='xx-large')
    plt.xticks(fontsize=20)
    plt.yticks(fontsize=20)
    plt.savefig('offpolicy_compare_coverage.png', bbox_inches='tight')
    plt.close()
